##***************************************************
## R source code for computing the denoised image and its gradient
## Created by Myriam Vimond
## Updated 27 July 2016
##***************************************************

## select.bandwidth()-------------------------------------
##    leave-one-out cross-validation for bandwidth selection with minimum MSE as objective criterion
## INPUT :
##      data = grayImage
##      ngrid = size of a grid of inverse of bandwidth for the cross-validation procedure
## OUTPUT : length two vector of inverse of bandwidths (expressed as a pixel number)
select.bandwidth <- function(data,ngrid){
  n1 <- dim(data)[1]
  n2 <- dim(data)[2]
  s1 <- 2*(n1+1)
  s2 <- 2*(n2+1)
  lmbd <- 1/1.1   
  inv_bdth <- matrix(c(s1,s2), nrow=ngrid^2, ncol=2, byrow=TRUE) * 
    matrix(c(rep(lmbd^(0:(ngrid-1)),ngrid),
             rep(lmbd^(0:(ngrid-1)),each=ngrid)),
           ncol=2)   
  inv_bdth <- localpoly.cv.grayImage(data, 
                                 bandwidth=inv_bdth)  # inverse of bandwidths (2*h matrix)
  inv_bdth
}

## localpoly.reg.grayImage()-------------------------------------
##    Local polynomial estimation of the image and its gradient with a truncated Gaussian kernel which is truncated between -3 and 3 (pixels).
## INPUT :
##       data = matrix
##  bandwidth = inverse of bandwidths (vector of inverse of bandwidths)
## OUTPUT : liste
##      F = denoised image 
##  gradF = list of the gradient component estimates
##          gF1 = estimated partial derivative of F with respect to the columns
##          gF2 = estimated partial derivative of F with respect to the rows
## REFERENCE :
##    require(NonpModelCheck)
localpoly.reg.grayImage <- function(data, 
                                    bandwidth = 0){
  ## image size
  dimension <- dim(data)
  n1 <- dimension[1]
  n2 <- dimension[2]
  
  ## initialization
  F <- matrix(0, nrow=n1, ncol=n2)
  
  gF1 <- matrix(0, nrow=n1, ncol=n2)
  gF2 <- matrix(0, nrow=n1, ncol=n2)

  ## size in pixel unit of kernel support 
  h <- 3
  ## inverse of bandwidth
  bdth2 = bandwidth[2]
  bdth1 = bandwidth[1]
  ## kernel filter
  Kgrid <- grille(2*h+1,2*h+1)
  Kgrid2 <- (Kgrid$dom2*(2*h+2)-(h+1))/(n2+1)
  Kgrid1 <- (Kgrid$dom1*(2*h+2)-(h+1))/(n1+1)
  K <- dnorm(Kgrid2,0,1/bdth2)*dnorm(Kgrid1,0,1/bdth1) 

  ## call .C function
  switch(Sys.info()[['sysname']],
         # Windows= {dyn.load("./lib/gkv-local-polynomials.dll")},
         Windows= {dyn.load(file.path(workDir,'lib','gkv-local-polynomials.dll',fsep = .Platform$file.sep))},
         # Linux = {dyn.load("./lib/gkv-local-polynomials-linux.so")},
         Linux = {dyn.load(file.path(workDir,'lib','gkv-local-polynomials-linux.so',fsep = .Platform$file.sep))},
         # Darwin = {dyn.load("./lib/gkv-local-polynomials-macosx.so")})
         Darwin = {dyn.load(file.path(workDir,'lib','gkv-local-polynomials-macosx.so',fsep = .Platform$file.sep))})
  out <- .C("localPolynomials",
            data=as.double(data),
            ny=as.integer(n1),
            nx=as.integer(n2),
            K=as.double(K),
            F=as.double(F),
            gFx=as.double(gF1),
            gFy=as.double(gF2))
  F <- matrix(out$F, n1, n2)
  gF1 <- matrix(out$gFx, n1, n2)
  gF2 <- matrix(out$gFy, n1, n2)
  
  ## RETURN
  list(F=F, gradF=list(gF1=gF1, gF2=gF2)) 
}

## localpoly.cv.grayImage()-------------------------------------
##    Cross-validation by using local polynomials of degree 1
## INPUT :
##       data = grayImage
##  bandwidth = inverse of bandwidths (vector or matrix)
## OUTPUT : bandwidth (vector)
## REFERENCE :
##    require(NonpModelCheck)
localpoly.cv.grayImage <- function(data, bandwidth){
  
  ## size of image
  dimension <- dim(data)
  n1 <- dimension[1]
  n2 <- dimension[2]
  
  ## initialization
  F <- matrix(0, nrow=n1, ncol=n2)

    gF1 <- matrix(0, nrow=n1, ncol=n2)
    gF2 <- matrix(0, nrow=n1, ncol=n2)
  
  nb <- NROW(bandwidth)
  cv <- numeric(nb)
  
  for(b in 1:nb){
    ## number of left or right pixels to the center of kernel support 
    h <- 3
    ## inverse of bandwidth
    bdth2 = bandwidth[b,2]
    bdth1 = bandwidth[b,1]
    ## kernel filter
    Kgrid <- grille(2*h+1,2*h+1)
    Kgrid2 <- (Kgrid$dom2*(2*h+2)-(h+1))/(n2+1)
    Kgrid1 <- (Kgrid$dom1*(2*h+2)-(h+1))/(n1+1)
    K <- dnorm(Kgrid2,0,1/bdth2)*dnorm(Kgrid1,0,1/bdth1) 
  
    ## call .C function
    switch(Sys.info()[['sysname']],
           # Windows= {dyn.load("./lib/gkv-local-polynomials.dll")},
           Windows= {dyn.load(file.path(workDir,'lib','gkv-local-polynomials-cv.dll',fsep = .Platform$file.sep))},
           # Linux = {dyn.load("./lib/gkv-local-polynomials-linux.so")},
           Linux = {dyn.load(file.path(workDir,'lib','gkv-local-polynomials-cv-linux.so',fsep = .Platform$file.sep))},
           # Darwin = {dyn.load("./lib/gkv-local-polynomials-macosx.so")})
           Darwin = {dyn.load(file.path(workDir,'lib','gkv-local-polynomials-cv-macosx.so',fsep = .Platform$file.sep))})
    out <- .C("localPolynomialscv",
            data=as.double(data),
            ny=as.integer(n1),
            nx=as.integer(n2),
            K=as.double(K),
            F=as.double(F),
            gFx=as.double(gF1),
            gFy=as.double(gF2))
    F <- matrix(out$F, n1, n2)
    cv[b] <- mean((data-F)^2)
  }
  bhat <- which.min(cv)  
    
  ## RETURN
  bandwidth[bhat,]
}




## finitediff()-------------------------------------
##    Gradient estimation by finite differences
## INPUT :
##      data = matrix
## OUTPUT : liste
##  gradF = list of the gradient component estimates
##          gF1 = estimated partial derivative of F with respect to the columns
##          gF2 = estimated partial derivative of F with respect to the rows
finitediff <- function(data, degree=1, method, bandwidth){
  ## size of data
  n1 <- dim(data)[1] 
  n2 <- dim(data)[2]
  ## width of discretisation grid
  dx1 = 1/(n1)  
  dx2 = 1/(n2)  
  F <- data
  ## Double difference filter
  gF1 <- matrix(0, nrow=n1, ncol=n2)
  gF2 <- matrix(0, nrow=n1, ncol=n2)
  ## boundary
  gF1[1,] = (F[2,] - F[1,])/dx1 
  gF2[,1] = (F[,2] - F[,1])/dx2
  gF1[n1,] = (F[n1,] - F[(n1-1),])/dx1 
  gF2[,n2] = (F[,n2] - F[,(n2-1)])/dx2
  ##
  gF1[2:(n1-1),] = (F[3:n1,] - F[1:(n1-2),])/(2*dx1)
  gF2[,2:(n2-1)] = (F[,3:n2] - F[,1:(n2-2)])/(2*dx2)   
  
  ## RETURN
  list(F=F,gF2=gF2,gF1=gF1)
}


