##***************************************************
## R source code for implementation of the estimation procedure of image reflectance and illumination
## Created by Myriam Vimond
## Updated 12 Nov 2014
##***************************************************

## load gkv functions
source('./lib/fnct02-gkv-utils.R')
source('./lib/fnct03-gkv-grad.R')

## illumination()-------------------------------------
##    Procedure for the estimation of image reflectance and illumination
## INPUT :
##      data = grayImage
##     cData = number of graylevels, 
##           "bitmap" for two levels
##           "8bits" for 256 levels
##           "10bits" for 1024 levels
##           "12bits" for 4096 levels
##    degree = degree of the polynomial basis of smooth space L
##    method = estimation method for F
##              "gaussian" : local polynomials with gaussian truncated kernel
##              "direct"   : no smoothing technique
##    ngrid = if "bandwidth" is missing, then "bandwidth" is selected by cross-validation with a grid-size equal to ngrid*ngrid
## bandwidth = inverse of bandwidth (needed for the local polynomial procedure)
## OUTPUT : list
##     F  = denoised image 
##     R  = reflectance estimation
##     L  = illumination estimation
##  gamma = vector of illumination coefficients 
illumination <- function(data, degree, method="gaussian", bandwidth, cData="8bits", ngrid=10){
  
  ## contraction needed for log conversion
  data <- rescale(data,cData="8bits")
  ## size of image
  n1 <- dim(data)[1] 
  n2 <- dim(data)[2]
  n  <- n1*n2
  ## coordinates of pixel grid
  grid <- grille(n1,n2)
  x1 <- grid$x1 
  x2 <- grid$x2 
  dom1 <- grid$dom1
  dom2 <- grid$dom2
  
  ## estimation of F ang grad(log(F))
   if(method == "direct"){
      outgrad <- finitediff(data)
      F <- outgrad$F
      gf2 <- outgrad$gF2/F 
      gf1 <- outgrad$gF1/F
      }
   else{
      if(missing(bandwidth)){
      ## 'missing()' is used to test whether a value was specified as an argument
      bandwidth <- select.bandwidth(data,ngrid)
    }
    ll <- localpoly.reg.grayImage(data, bandwidth)
    F <- ll$F
    gf2 <- ll$gradF$gF2/F 
    gf1 <- ll$gradF$gF1/F  
  }
  gf2 <- matrix(gf2,nrow=n,ncol=1)
  gf1 <- matrix(gf1,nrow=n,ncol=1)
  
  ## polynomial basis
  deg1 <- degree[1]; deg2 <- degree[2];
  gp2 <- matrix(0, nrow=n, ncol=(deg2+1)*(deg1+1));
  gp1 <- matrix(0, nrow=n, ncol=(deg2+1)*(deg1+1));
  for(d2 in 0:deg2){
    for(d1 in 0:deg1){
      d <- d1+1+(deg1+1)*(d2)
      gp2[,d] <- matrix(d2*exp(log(dom2)*(d2-1))*exp(log(dom1)*(d1)),nrow=n,ncol=1)
      gp1[,d] <- matrix(d1*exp(log(dom2)*(d2))*exp(log(dom1)*(d1-1)),nrow=n,ncol=1)
    }
  } 
      
  ## Scalar product between grad(f) and grad(polynomes)
  ## Be careful : we have to remove the degree (0,0) coefficient
  b <- matrix(0, nrow=(deg2+1)*(deg1+1), ncol=1)
  b <- t(gp2)%*%gf2 + t(gp1)%*%gf1 
  b = b[2:((deg2+1)*(deg1+1)),]
  
  ## Gram matrix of grad(polynomes)
  A <- matrix(0, nrow=(deg2+1)*(deg1+1), ncol=(deg2+1)*(deg1+1))
  A <- t(gp2)%*%gp2 + t(gp1)%*%gp1
  A <- A[2:((deg2+1)*(deg1+1)),2:((deg2+1)*(deg1+1))]
  
  ## Estimation of gamma
  gamma <- solve(A,b)
  ## Illumination estimation 
  gamma <- matrix(c(0,gamma), nrow=(deg1+1), ncol=(deg2+1))
  lumi <- matrix(0, nrow=n1, ncol=n2);
  for(d2 in 1:(deg2+1)){
    for(d1 in 1:(deg1+1)){
      d <- d1+(deg1+1)*(d2-1)
      lumi <- lumi + gamma[d1,d2]*(exp(log(dom2)*(d2-1)+log(dom1)*(d1-1)))
    }
  }
  L <- exp(lumi)/mean(exp(lumi))
  ## Reflectance estimation 
  R =  F/L
  
  ## RETURN
  if(method=="gaussian"){
    pars <- list(illumination=list(model="the log-illumination is a polynomial",degree=degree), 
                 gradient=list(smoothing="local polynomials",
                               degree =1,
                               bandwidth=1/bandwidth,
                               family="trunc gaussian"))
  }else{
    pars <- list(illumination=list(model="the log-illumination is a polynomial",degree=degree), 
                 gradient=list(smoothing="double difference"))
  }

  list(F=F, R=R, L=L, gamma=gamma, pars=pars, data.size=c(n1,n2))
}

